"""Top-level package for alfasim-sdk."""

import pluggy

from alfasim_sdk._internal import version
from alfasim_sdk._internal.alfasim_sdk_utils import get_metadata
from alfasim_sdk._internal.units import register_units

# Package information
__author__ = "ESSS"
__email__ = "foss@esss.co"
__version__ = version.__version__

# Register new categories and units used from CaseDescription
register_units()

# Register ALFASim-SDK Hooks
hookimpl = pluggy.HookimplMarker("ALFAsim")


def get_alfasim_sdk_api_path():
    """
    Return the directory that contains the alfasim_sdk_api with the header files
    """
    from pathlib import Path

    from alfasim_sdk._internal import hook_specs

    return str(Path(hook_specs.__file__).parents[1])


# CLI:
# ALFACase: Utilities
from alfasim_sdk._internal.alfacase.alfacase import (
    convert_alfacase_to_description,
    convert_description_to_alfacase,
    generate_alfacase_file,
)
from alfasim_sdk._internal.alfacase.alfatable import (
    generate_alfatable_file,
    load_pvt_model_table_parameters_description_from_alfatable,
)

# ALFACase: Descriptions
from alfasim_sdk._internal.alfacase.case_description import (
    AlfasimVersionInfo,
    AnnulusDescription,
    AnnulusEquipmentDescription,
    BipDescription,
    CaseDescription,
    CaseOutputDescription,
    CasingDescription,
    CasingSectionDescription,
    CombinedFluidDescription,
    CompositionalFluidDescription,
    CompositionDescription,
    CompressorEquipmentDescription,
    CompressorPressureTableDescription,
    ControllerInputSignalPropertiesDescription,
    ControllerNodePropertiesDescription,
    ControllerOutputSignalPropertiesDescription,
    ControllerTrendDescription,
    CvTableDescription,
    EnvironmentDescription,
    EnvironmentPropertyDescription,
    EquipmentDescription,
    EquipmentTrendDescription,
    FetkovichIPRDescription,
    ForchheimerIPRDescription,
    FormationDescription,
    FormationLayerDescription,
    GasLiftValveEquipmentDescription,
    GlobalTrendDescription,
    HeatSourceEquipmentDescription,
    HeavyComponentDescription,
    InitialConditionsDescription,
    InitialPressuresDescription,
    InitialTemperaturesDescription,
    InitialTracersMassFractionsDescription,
    InitialVelocitiesDescription,
    InitialVolumeFractionsDescription,
    InternalNodePropertiesDescription,
    InternalReferencePluginTableColumn,
    IPRCurveDescription,
    IPRModelsDescription,
    LeakEquipmentDescription,
    LengthAndElevationDescription,
    LightComponentDescription,
    LinearIPRDescription,
    MassSourceEquipmentDescription,
    MassSourceNodePropertiesDescription,
    MaterialDescription,
    NodeDescription,
    NumericalOptionsDescription,
    OpenHoleDescription,
    OverallPipeTrendDescription,
    PackerDescription,
    PhysicsDescription,
    PigEquipmentDescription,
    PipeDescription,
    PipeSegmentsDescription,
    PluginDescription,
    PositionalPipeTrendDescription,
    PressureContainerDescription,
    PressureNodePropertiesDescription,
    ProfileDescription,
    ProfileOutputDescription,
    PumpEquipmentDescription,
    PvtModelCombinedDescription,
    PvtModelCompositionalDescription,
    PvtModelConstantPropertiesDescription,
    PvtModelCorrelationDescription,
    PvtModelPhTableParametersDescription,
    PvtModelPtTableParametersDescription,
    PvtModelsDescription,
    ReferencedPressureContainerDescription,
    ReferencedTemperaturesContainerDescription,
    ReferencedTracersMassFractionsContainerDescription,
    ReferencedVelocitiesContainerDescription,
    ReferencedVolumeFractionsContainerDescription,
    ReservoirInflowEquipmentDescription,
    SeparatorNodePropertiesDescription,
    SeparatorTrendDescription,
    SpeedCurveDescription,
    SurgeVolumeOptionsDescription,
    TableIPRDescription,
    TablePumpDescription,
    TemperaturesContainerDescription,
    TimeOptionsDescription,
    TracerModelConstantCoefficientsDescription,
    TracerReferencePluginTableColumn,
    TracersDescription,
    TracersMassFractionsContainerDescription,
    TrendsOutputDescription,
    TubingDescription,
    ValveEquipmentDescription,
    VelocitiesContainerDescription,
    VogelIPRDescription,
    VolumeFractionsContainerDescription,
    WallDescription,
    WallLayerDescription,
    WellDescription,
    XAndYDescription,
)
from alfasim_sdk._internal.cli import console_main

# Constants
from alfasim_sdk._internal.constants import (
    BUBBLE_FIELD,
    DEFAULT_TEMPERATURE_IN_K,
    DROPLET_FIELD,
    EXTRAS_REQUIRED_VERSION_KEY,
    FLUID_GAS,
    FLUID_OIL,
    FLUID_PHASE_NAMES,
    FLUID_WATER,
    GAS_FIELD,
    GAS_LAYER,
    GAS_PHASE,
    OIL_FIELD,
    OIL_LAYER,
    OIL_PHASE,
    SOLID_PHASE,
    WATER_DROPLET_IN_OIL_FIELD,
    WATER_FIELD,
    WATER_LAYER,
    WATER_PHASE,
    CompressorSpeedType,
    ControllerType,
    CorrelationPackage,
    CorrelationPackageType,
    CorrelationsGasViscosity,
    CorrelationsOilViscosity,
    CorrelationsSurfaceTension,
    DrainageRateMode,
    EmulsionDropletSizeModelType,
    EmulsionInversionPointModelType,
    EmulsionModelType,
    EmulsionRelativeViscosityModelType,
    EnergyModel,
    EnergyModelPrimaryVariable,
    EquationOfStateType,
    EquipmentAttachmentLocation,
    EvaluationStrategyType,
    FlashModel,
    FlowDirection,
    FlowPatternModel,
    FluidMaterialConvectionCorrelation,
    ForchheimerCoefficientsOption,
    GasLiftValveOpeningType,
    HydrodynamicModelType,
    InitialConditionStrategyType,
    InterpolationType,
    LeakLocation,
    LeakModel,
    LeakType,
    MassInflowSplitType,
    MassSourceType,
    MaterialType,
    MultiInputType,
    NodeCellType,
    NonlinearSolverType,
    OutputAttachmentLocation,
    PigRoutingMode,
    PigTrappingMode,
    PipeEnvironmentHeatTransferCoefficientModelType,
    PipeThermalModelType,
    PipeThermalPositionInput,
    PumpThermalEfficiencyModel,
    PumpType,
    PumpViscosityModel,
    PVTCompositionalViscosityModel,
    SeparatorGeometryType,
    SimulationModeType,
    SimulationRegimeType,
    SolidsModelType,
    SurfaceTensionType,
    SurgeVolumeTimeMode,
    TableInputType,
    TracerModelType,
    ValveOpeningType,
    ValveType,
    WellConnectionPort,
    WellIndexPhaseType,
)

# Plugins: Context
from alfasim_sdk._internal.context import Context

# Plugins: Layouts imports
from alfasim_sdk._internal.layout import group, tab, tabs

# Plugins: Models imports
from alfasim_sdk._internal.models import container_model, data_model

# Plugins: Status imports
from alfasim_sdk._internal.status import ErrorMessage, WarningMessage

# Plugins: Types for configure_fields hook
# Plugins: Types for configure_phases hook
# Plugins: Types for configure_layers hook
# Plugins: Types  for UI customization
from alfasim_sdk._internal.types import (
    AddField,
    AddLayer,
    AddPhase,
    BaseField,
    Boolean,
    Enum,
    FileContent,
    MultipleReference,
    Quantity,
    Reference,
    String,
    Table,
    TableColumn,
    TracerType,
    UpdateLayer,
    UpdatePhase,
)

# Plugins: Type for alfasim_get_additional_variables hook
# Plugins: Constants used on SecondaryVariable
from alfasim_sdk._internal.variables import (
    Location,
    Scope,
    SecondaryVariable,
    Type,
    Visibility,
)

__all__ = [
    "AddField",
    "AddLayer",
    "AddPhase",
    "AnnulusDescription",
    "AnnulusEquipmentDescription",
    "BUBBLE_FIELD",
    "BaseField",
    "BipDescription",
    "Boolean",
    "AlfasimVersionInfo",
    "CaseDescription",
    "CaseOutputDescription",
    "CasingDescription",
    "CasingSectionDescription",
    "CompositionDescription",
    "CompressorEquipmentDescription",
    "CompressorPressureTableDescription",
    "CompressorSpeedType",
    "Context",
    "ControllerInputSignalPropertiesDescription",
    "ControllerNodePropertiesDescription",
    "ControllerOutputSignalPropertiesDescription",
    "ControllerTrendDescription",
    "ControllerType",
    "CorrelationPackage",
    "CorrelationPackageType",
    "CorrelationsOilViscosity",
    "CorrelationsGasViscosity",
    "CorrelationsSurfaceTension",
    "CvTableDescription",
    "DEFAULT_TEMPERATURE_IN_K",
    "DROPLET_FIELD",
    "DrainageRateMode",
    "EXTRAS_REQUIRED_VERSION_KEY",
    "EmulsionModelType",
    "EmulsionRelativeViscosityModelType",
    "EmulsionDropletSizeModelType",
    "EmulsionInversionPointModelType",
    "EnergyModel",
    "EnergyModelPrimaryVariable",
    "Enum",
    "EnvironmentDescription",
    "EnvironmentPropertyDescription",
    "EquationOfStateType",
    "EquipmentAttachmentLocation",
    "EquipmentDescription",
    "EquipmentTrendDescription",
    "ErrorMessage",
    "EvaluationStrategyType",
    "FLUID_GAS",
    "FLUID_OIL",
    "FLUID_PHASE_NAMES",
    "FLUID_WATER",
    "FileContent",
    "FlashModel",
    "FlowDirection",
    "FlowPatternModel",
    "FluidMaterialConvectionCorrelation",
    "ForchheimerIPRDescription",
    "ForchheimerCoefficientsOption",
    "CompositionalFluidDescription",
    "CombinedFluidDescription",
    "FormationDescription",
    "FormationLayerDescription",
    "GAS_FIELD",
    "GAS_LAYER",
    "GAS_PHASE",
    "GasLiftValveEquipmentDescription",
    "GasLiftValveOpeningType",
    "GlobalTrendDescription",
    "HeatSourceEquipmentDescription",
    "HeavyComponentDescription",
    "HydrodynamicModelType",
    "IPRCurveDescription",
    "IPRModelsDescription",
    "InitialConditionStrategyType",
    "InitialConditionsDescription",
    "InitialPressuresDescription",
    "InitialTemperaturesDescription",
    "InitialTracersMassFractionsDescription",
    "InitialVelocitiesDescription",
    "InitialVolumeFractionsDescription",
    "InternalNodePropertiesDescription",
    "InternalReferencePluginTableColumn",
    "InterpolationType",
    "LeakEquipmentDescription",
    "LeakLocation",
    "LeakModel",
    "LeakType",
    "LengthAndElevationDescription",
    "LightComponentDescription",
    "LinearIPRDescription",
    "Location",
    "MassInflowSplitType",
    "MassSourceEquipmentDescription",
    "MassSourceNodePropertiesDescription",
    "MassSourceType",
    "MaterialDescription",
    "MaterialType",
    "MultiInputType",
    "MultipleReference",
    "NodeCellType",
    "NodeDescription",
    "NonlinearSolverType",
    "NumericalOptionsDescription",
    "OIL_FIELD",
    "OIL_LAYER",
    "OIL_PHASE",
    "OpenHoleDescription",
    "OutputAttachmentLocation",
    "OverallPipeTrendDescription",
    "PVTCompositionalViscosityModel",
    "PackerDescription",
    "PhysicsDescription",
    "PigEquipmentDescription",
    "PigRoutingMode",
    "PigTrappingMode",
    "PipeDescription",
    "PipeEnvironmentHeatTransferCoefficientModelType",
    "PipeSegmentsDescription",
    "PipeThermalModelType",
    "PipeThermalPositionInput",
    "PluginDescription",
    "PositionalPipeTrendDescription",
    "PressureContainerDescription",
    "PressureNodePropertiesDescription",
    "ProfileDescription",
    "ProfileOutputDescription",
    "PumpEquipmentDescription",
    "PumpType",
    "PumpViscosityModel",
    "PumpThermalEfficiencyModel",
    "PvtModelCombinedDescription",
    "PvtModelCompositionalDescription",
    "PvtModelCorrelationDescription",
    "PvtModelConstantPropertiesDescription",
    "PvtModelPtTableParametersDescription",
    "PvtModelPhTableParametersDescription",
    "PvtModelsDescription",
    "Quantity",
    "Reference",
    "ReferencedPressureContainerDescription",
    "ReferencedTemperaturesContainerDescription",
    "ReferencedTracersMassFractionsContainerDescription",
    "ReferencedVelocitiesContainerDescription",
    "ReferencedVolumeFractionsContainerDescription",
    "ReservoirInflowEquipmentDescription",
    "SOLID_PHASE",
    "Scope",
    "SecondaryVariable",
    "SeparatorGeometryType",
    "SeparatorNodePropertiesDescription",
    "SeparatorTrendDescription",
    "SimulationModeType",
    "SimulationRegimeType",
    "SolidsModelType",
    "SpeedCurveDescription",
    "String",
    "SurfaceTensionType",
    "SurgeVolumeOptionsDescription",
    "SurgeVolumeTimeMode",
    "Table",
    "TableColumn",
    "TableIPRDescription",
    "TableInputType",
    "TablePumpDescription",
    "TemperaturesContainerDescription",
    "TimeOptionsDescription",
    "TracerModelConstantCoefficientsDescription",
    "TracerModelType",
    "TracerReferencePluginTableColumn",
    "TracerType",
    "TracersDescription",
    "TracersMassFractionsContainerDescription",
    "TrendsOutputDescription",
    "TubingDescription",
    "Type",
    "UpdateLayer",
    "UpdatePhase",
    "FetkovichIPRDescription",
    "ValveEquipmentDescription",
    "ValveOpeningType",
    "ValveType",
    "VelocitiesContainerDescription",
    "Visibility",
    "VogelIPRDescription",
    "VolumeFractionsContainerDescription",
    "WATER_DROPLET_IN_OIL_FIELD",
    "WATER_FIELD",
    "WATER_LAYER",
    "WATER_PHASE",
    "WallDescription",
    "WallLayerDescription",
    "WarningMessage",
    "WellConnectionPort",
    "WellDescription",
    "WellIndexPhaseType",
    "XAndYDescription",
    "console_main",
    "container_model",
    "convert_alfacase_to_description",
    "convert_description_to_alfacase",
    "data_model",
    "generate_alfacase_file",
    "generate_alfatable_file",
    "get_metadata",
    "group",
    "load_pvt_model_table_parameters_description_from_alfatable",
    "tab",
    "tabs",
]
